/*
  CRF++ -- Yet Another CRF toolkit

  $Id: feature_index.h,v 1.15 2006/03/28 18:15:59 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#ifndef _CRFPP_FEATURE_INDEX_H
#define _CRFPP_FEATURE_INDEX_H

#include <vector>
#include <map>
#include <iostream>
#include "common.h"
#include "scoped_ptr.h"
#include "feature_cache.h"
#include "path.h"
#include "node.h"
#include "freelist.h"
#include "mmap.h"
#include "darts.h"

namespace CRFPP
{
  class TaggerImpl;
  
  class FeatureIndex
  {
  protected:
    unsigned int maxid;
    double *alpha;
    float  *alpha_float;
    unsigned int xsize_;
    size_t thread_num;          
    FeatureCache       feature_cache;
    std::vector<char*> unigram_templs;
    std::vector<char*> bigram_templs;
    std::vector<char*>	y;
    FreeList<char>  char_freelist;
    scoped_array< FreeList<Path> > path_freelist;
    scoped_array< FreeList<Node> > node_freelist;
    virtual int getID (const char *) = 0;
    string_buffer _what;

  public:
    static const unsigned int version = MODEL_VERSION;

    size_t size	 () const { return maxid; }
    size_t xsize () const { return xsize_; }
    size_t ysize () const { return y.size(); }
    const char* yname (size_t i) const { return y[i]; }
    void   set_alpha (double *_alpha) { alpha = _alpha; }

    char *strdup (const char *);
    void calcCost (Node *);
    void calcCost (Path *);

    bool buildFeatures   (TaggerImpl &);
    void rebuildFeatures (TaggerImpl &);

    const char* what() { return _what.c_str(); }

    virtual bool open (const char*, const char*) = 0;
    virtual void clear () = 0;
     
    void init() {
      path_freelist.reset(new FreeList<Path> [thread_num]);
      node_freelist.reset(new FreeList<Node> [thread_num]);
      for (size_t i = 0; i < thread_num; ++i) {
	path_freelist[i].set_size(8192 * 16);
	node_freelist[i].set_size(8192);
      }
    }

    FeatureIndex(): maxid(0), alpha(0), alpha_float(0),
                    thread_num(1), char_freelist(8192) {}
    virtual ~FeatureIndex () {};
  };

  class EncoderFeatureIndex: public FeatureIndex {
  private:
    std::map <std::string, std::pair<int, unsigned int> > dic;
    int getID (const char *);
    bool openTemplate (const char *);
    bool openTagSet   (const char *);
  public:
    EncoderFeatureIndex(size_t n) { thread_num = n; init(); }
    bool open(const char*, const char*);
    bool save (const char *, bool);
    bool convert (const char *, const char*);
    void clear ();
    void shrink (size_t) ;
  };

  class DecoderFeatureIndex: public FeatureIndex {
  private:
    Mmap <char> mmap;
    Darts::DoubleArray da;
    int getID (const char *);
   public:
    DecoderFeatureIndex() { thread_num = 1; init(); }
    bool open (const char *, const char *);
    void clear ();
  };
}

#endif
