/*
  MeCab -- Yet Another Part-of-Speech and Morphological Analyzer
 
  $Id: lbfgs.h,v 1.8 2006/03/19 17:38:41 taku Exp $;

  Copyright (C) 2001-2005 Taku Kudo <taku@chasen.org>
  Copyright (C) 2004-2005 Nippon Telegraph and Telephone Corporation

  This is free software with ABSOLUTELY NO WARRANTY.
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef _CRFPP_LBFGS_H
#define _CRFPP_LBFGS_H

#include <vector>
#include <string>
#include "common.h"

extern "C" {
  extern void lbfgs(int* n, int* m, double* x, double* f, double* g, 
		    int* diagco, double* diag, int* iprint, double* eps, 
		    double* xtol, double* w, int* iflag);
}

namespace CRFPP {

  class LBFGS {

  private:
    int n;
    int m;
    int iflag;
    string_buffer _what;
    std::vector <double> diag;
    std::vector <double> w;

  public:
    // these valuables are fixed.
    LBFGS(): n(0), m(5), iflag(0) {};
    ~LBFGS() {};

    const char* what () { return _what.c_str(); }

    int init (int _n, int _m)
    {
      n = _n;
      m = _m;
      w.resize (n * (2 * m + 1) + 2 * m);
      diag.resize (n);
      return 0;
    }
  
    int optimize (double *x, double *f, double *g)   // vector, object, gradient
    {
       int iprint[] = {-1, 0};
       double eta  = 1e-7;
       double xtol = 1e-7;
       int diagco = 0;
       lbfgs(&n, &m, x, f, g, &diagco, &diag[0], iprint, &eta, &xtol, &w[0], &iflag);
       
       if (iflag < 0) {
	 WHAT("routine stops with unexpected error");
	 return -1;
       } 

       if (iflag == 0) return 0; // terminate
       
       return 1; // evaluate next f and g
    }
  };
}

#endif
