/*
  CRF++ -- Yet Another CRF toolkit

  $Id: timer.h,v 1.6 2006/03/20 12:52:52 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#ifndef _CRFPP_TIMER_H
#define _CRFPP_TIMER_H

#include <ctime>
#include <iostream>
#include <string>
#include <limits>

/* COPY FROM Boost::timer */

// namespace std { using ::clock_t; using ::clock; }

namespace CRFPP {

  class timer
  {
  public:
    timer() { _start_time = std::clock(); }
    void   restart() { _start_time = std::clock(); }
    double elapsed() const
    { return  double(std::clock() - _start_time) / CLOCKS_PER_SEC; }

    double elapsed_max() const
    {
      return (double(std::numeric_limits<std::clock_t>::max())
	      - double(_start_time)) / double(CLOCKS_PER_SEC);
    }

    double elapsed_min() const
    { return double(1)/double(CLOCKS_PER_SEC); }

  private:
    std::clock_t _start_time;
  };

  class progress_timer : public timer
  {

  public:
    explicit progress_timer( std::ostream & os = std::cout ) : m_os(os) {}
    ~progress_timer()
    {
	std::istream::fmtflags old_flags = m_os.setf( std::istream::fixed,
						      std::istream::floatfield );
	std::streamsize old_prec = m_os.precision( 2 );
	m_os << elapsed() << " s\n" << std::endl;
	m_os.flags( old_flags );
	m_os.precision( old_prec );
    }

  private:
    std::ostream & m_os;
  };
}

#endif
